@@ -13,4 +13,5 @@ from .separable_convolution import *
 from .reshaping import *
 from .attention import *
 from .pooling import *
-from .reciprocal import *
+from .recorders import *
+from .rescaling import *
