@@ -36,7 +36,7 @@ class QuantizedPermute(keras.layers.Permute):
             of the input.
     """
 
-    def call(self, inputs, training=None):
+    def call(self, inputs):
         # raise an error if the inputs are not FixedPoint
         if not isinstance(inputs, FixedPoint):
             raise TypeError(f"QuantizedPermute only accepts FixedPoint inputs.\
@@ -49,4 +49,4 @@ class QuantizedPermute(keras.layers.Permute):
         # Transpose only the values
         transposed_values = super().call(inputs.values)
         # Return a FixedPoint with the modified values
-        return FixedPoint(transposed_values, inputs.frac_bits, inputs.value_bits)
+        return FixedPoint(transposed_values, inputs.value_bits, inputs.frac_bits)
