@@ -17,7 +17,7 @@
 import tensorflow as tf
 import keras
 
-from ...tensors import FixedPoint
+from ...tensors import QTensor, FixedPoint
 
 
 __all__ = ["QuantizedReshape"]
@@ -32,16 +32,19 @@ class QuantizedReshape(keras.layers.Reshape):
             dimension (batch size).
     """
 
-    def call(self, inputs, training=None):
-        # raise an error if the inputs are not FixedPoint
-        if not isinstance(inputs, FixedPoint):
-            raise TypeError(f"QuantizedReshape only accepts FixedPoint inputs.\
+    def call(self, inputs):
+        # raise an error if the inputs are not QTensor
+        if not isinstance(inputs, QTensor):
+            raise TypeError(f"QuantizedReshape only accepts QTensor inputs.\
                              Receives {type(inputs)} inputs.")
-        if not inputs.per_tensor:
-            # Different fractional-bits are defined for the last axis, so
-            # it must be preserved during the reshape
-            tf.assert_equal(self.target_shape[-1], tf.shape(inputs)[-1])
-        # Reshape the values
-        reshaped_values = super().call(inputs.values)
-        # Return a new reshaped FixedPoint
-        return FixedPoint(reshaped_values, inputs.frac_bits, inputs.value_bits)
+        # Return a new reshaped QTensor
+        result = tf.reshape(inputs, (tf.shape(inputs)[0],) + self.target_shape)
+        if not tf.executing_eagerly():
+            # Set the static shape for the result since it might lost during
+            # array_ops reshape, eg, some `None` dim in the result could be
+            # inferred.
+            if isinstance(result, FixedPoint):
+                result.values.set_shape(self.compute_output_shape(inputs.shape))
+            else:
+                result.fp.values.set_shape(self.compute_output_shape(inputs.shape))
+        return result
