@@ -24,8 +24,9 @@ from keras.utils import conv_utils
 from keras import backend
 
 from ..tensors import FixedPoint, MAX_BUFFER_BITWIDTH, QFloat
-from .calibrable import Calibrable, CalibrableVariable
-from .quantizers import deserialize_quantizer, WeightQuantizer, OutputQuantizer
+from .recorders import TensorRecorder
+from .quantizers import (WeightQuantizer, AlignedWeightQuantizer,
+                         OutputQuantizer)
 
 
 __all__ = ["QuantizedSeparableConv2D", "SeparableConv2DTranspose",
@@ -33,7 +34,7 @@ __all__ = ["QuantizedSeparableConv2D", "SeparableConv2DTranspose",
 
 
 @tf.keras.utils.register_keras_serializable()
-class QuantizedSeparableConv2D(Calibrable, SeparableConv2D):
+class QuantizedSeparableConv2D(SeparableConv2D):
     """ A separable convolutional layer that operates on quantized inputs and weights.
 
     Args:
@@ -54,30 +55,29 @@ class QuantizedSeparableConv2D(Calibrable, SeparableConv2D):
         super().__init__(*args, **kwargs)
         self.quant_config = quant_config
 
-        self.out_quantizer = deserialize_quantizer(
-            self.quant_config, "output_quantizer", OutputQuantizer, False)
+        out_quant_cfg = quant_config.get("output_quantizer", False)
+        if out_quant_cfg:
+            self.out_quantizer = OutputQuantizer(
+                name="output_quantizer", **out_quant_cfg)
+        else:
+            self.out_quantizer = None
 
         # Separable layer has two weights quantizers to handle different max values
-        self.dw_weight_quantizer = deserialize_quantizer(
-            self.quant_config, "dw_weight_quantizer", WeightQuantizer, True)
+        self.dw_weight_quantizer = WeightQuantizer(
+            name="dw_weight_quantizer", **quant_config["dw_weight_quantizer"])
         if self.dw_weight_quantizer.get_config()['axis'] != "per-tensor":
             raise ValueError("Separable depthwise weights must be quantized per tensor")
-        self.pw_weight_quantizer = deserialize_quantizer(
-            self.quant_config, "pw_weight_quantizer", WeightQuantizer, True)
 
+        self.pw_weight_quantizer = WeightQuantizer(
+            name="pw_weight_quantizer", **quant_config["pw_weight_quantizer"])
         if self.use_bias:
-            self.bias_quantizer = deserialize_quantizer(
-                self.quant_config, "bias_quantizer", WeightQuantizer, True)
+            self.bias_quantizer = AlignedWeightQuantizer(
+                name="bias_quantizer", **quant_config["bias_quantizer"])
 
         self.buffer_bitwidth = self.quant_config.get("buffer_bitwidth", MAX_BUFFER_BITWIDTH) - 1
         assert self.buffer_bitwidth > 0, "The buffer_bitwidth must be a strictly positive integer."
 
-        if self.use_bias:
-            # Add objects that will store the shift values.
-            self.bias_shift = CalibrableVariable()
-            self.output_shift = CalibrableVariable()
-
-    def call(self, inputs, training=None):
+    def call(self, inputs):
         # raise an error if the inputs are not FixedPoint or tf.Tensor
         if not isinstance(inputs, (FixedPoint, tf.Tensor)):
             raise TypeError(f"QuantizedSeparableConv2D only accepts FixedPoint\
@@ -86,15 +86,15 @@ class QuantizedSeparableConv2D(Calibrable, SeparableConv2D):
         if isinstance(inputs, tf.Tensor):
             # Assume the inputs are integer stored as float, which is the only tf.Tensor
             # inputs that are allowed
-            inputs = FixedPoint(inputs, 0, 8)
+            inputs = FixedPoint(inputs, 8, 0)
 
         # Although the dephwise operation does not require it, we only accept inputs quantized
         # per-tensor to avoid increasing too much the fractional bits of the depthwise outputs.
         inputs.assert_per_tensor()
 
         # Quantize the weights
-        depthwise_kernel = self.dw_weight_quantizer(self.depthwise_kernel, training)
-        pointwise_kernel = self.pw_weight_quantizer(self.pointwise_kernel, training)
+        depthwise_kernel = self.dw_weight_quantizer(self.depthwise_kernel)
+        pointwise_kernel = self.pw_weight_quantizer(self.pointwise_kernel)
 
         inputs = inputs.promote(self.buffer_bitwidth)
         dw_outputs_q = backend.depthwise_conv2d(
@@ -113,17 +113,12 @@ class QuantizedSeparableConv2D(Calibrable, SeparableConv2D):
             data_format=conv_utils.convert_data_format(self.data_format, ndim=4))
 
         if self.use_bias:
-            bias = self.bias_quantizer(self.bias, training).promote(self.buffer_bitwidth)
-
-            # Align intermediate outputs and biases before adding them
-            outputs, shift = outputs.align(bias)
-            self.output_shift(shift)
-            bias, shift = bias.align(outputs)
-            self.bias_shift(shift)
+            # Quantize bias and align it on the outputs
+            bias = self.bias_quantizer(self.bias, outputs)
             outputs = tf.add(outputs, bias)
 
         if self.out_quantizer is not None:
-            outputs = self.out_quantizer(outputs, training)
+            outputs = self.out_quantizer(outputs)
         return outputs
 
     def get_config(self):
@@ -207,26 +202,29 @@ class SeparableConv2DTranspose(SeparableConv2D):
 
 
 @tf.keras.utils.register_keras_serializable()
-class QuantizedSeparableConv2DTranspose(Calibrable, SeparableConv2DTranspose):
+class QuantizedSeparableConv2DTranspose(SeparableConv2DTranspose):
     """ A transposed separable convolutional layer that operates on quantized inputs and weights.
 
     Args:
-        quant_config (dict, optional): the serialized quantization configuration.
-            Defaults to empty configuration.
+        quant_config (dict): the serialized quantization configuration.
     """
 
-    def __init__(self, *args, quant_config={}, **kwargs):
+    def __init__(self, *args, quant_config, **kwargs):
         super().__init__(*args, **kwargs)
         self.quant_config = quant_config
 
-        self.out_quantizer = deserialize_quantizer(
-            self.quant_config, "output_quantizer", OutputQuantizer, False)
+        out_quant_cfg = quant_config.get("output_quantizer", False)
+        if out_quant_cfg:
+            self.out_quantizer = OutputQuantizer(
+                name="output_quantizer", **out_quant_cfg)
+        else:
+            self.out_quantizer = None
 
         # Transposed Separable layer has two weights quantizers to handle different max values
-        self.dw_weight_quantizer = deserialize_quantizer(
-            self.quant_config, "dw_weight_quantizer", WeightQuantizer, True)
-        self.pw_weight_quantizer = deserialize_quantizer(
-            self.quant_config, "pw_weight_quantizer", WeightQuantizer, True)
+        self.dw_weight_quantizer = WeightQuantizer(
+            name="dw_weight_quantizer", **quant_config["dw_weight_quantizer"])
+        self.pw_weight_quantizer = WeightQuantizer(
+            name="pw_weight_quantizer", **quant_config["pw_weight_quantizer"])
 
         # Depthwise quantizer must be per-tensor
         if "axis" in self.quant_config["dw_weight_quantizer"]:
@@ -235,13 +233,17 @@ class QuantizedSeparableConv2DTranspose(Calibrable, SeparableConv2DTranspose):
                                  f"{self.quant_config['dw_weight_quantizer']['axis']}.")
 
         if self.use_bias:
-            self.bias_quantizer = deserialize_quantizer(
-                self.quant_config, "bias_quantizer", WeightQuantizer)
+            self.bias_quantizer = AlignedWeightQuantizer(
+                name="bias_quantizer", **quant_config["bias_quantizer"])
 
         self.buffer_bitwidth = self.quant_config.get("buffer_bitwidth", MAX_BUFFER_BITWIDTH) - 1
         assert self.buffer_bitwidth > 0, "The buffer_bitwidth must be a strictly positive integer."
-        self.intermediate_quantizer = deserialize_quantizer(
-            self.quant_config, "intermediate_quantizer", OutputQuantizer, False)
+        intermediate_quant_cfg = quant_config.get("intermediate_quantizer", False)
+        if out_quant_cfg:
+            self.intermediate_quantizer = OutputQuantizer(
+                name="intermediate_quantizer", **intermediate_quant_cfg)
+        else:
+            self.intermediate_quantizer = None
 
         if self.intermediate_quantizer:
             intermediate_quantizer_axis = self.intermediate_quantizer.get_config()['axis']
@@ -250,11 +252,9 @@ class QuantizedSeparableConv2DTranspose(Calibrable, SeparableConv2DTranspose):
                                  f"{intermediate_quantizer_axis}.")
 
         # Add objects that will store the shift values.
-        self.input_shift = CalibrableVariable()
-        self.pw_shift = CalibrableVariable()
-        self.bias_shift = CalibrableVariable()
+        self.input_shift = TensorRecorder()
 
-    def call(self, inputs, training=None):
+    def call(self, inputs):
         # raise an error if the inputs are not FixedPoint or tf.Tensor
         if not isinstance(inputs, (FixedPoint, tf.Tensor)):
             raise TypeError("QuantizedSeparableConv2DTranspose only accepts FixedPoint or "
@@ -265,8 +265,8 @@ class QuantizedSeparableConv2DTranspose(Calibrable, SeparableConv2DTranspose):
             # inputs that are allowed
             inputs = FixedPoint.quantize(inputs, 0, 8)
 
-        # Promote inputs to avoid saturation, then align them
-        inputs, shift = inputs.promote(self.buffer_bitwidth).align()
+        # Expand the inputs to a higher bitwidth to avoid saturation and align them
+        inputs, shift = inputs.expand(self.buffer_bitwidth)
         self.input_shift(shift)
 
         # Infer the dynamic output shape
@@ -283,6 +283,9 @@ class QuantizedSeparableConv2DTranspose(Calibrable, SeparableConv2DTranspose):
                                                     dilation=self.dilation_rate[1])
         output_shape = tf.stack((inputs_shape[0], out_height, out_width, 1))
 
+        # Quantize the depthwise kernels
+        depthwise_kernel = self.dw_weight_quantizer(self.depthwise_kernel)
+
         # Inputs and kernels must be transposed to have their channel dimension first because the
         # tf.vectorized_map call that follows will unpack them on dimension 0. The channel dimension
         # is virtually restored using expand_dims so that elements have the appropriate shape for
@@ -290,12 +293,9 @@ class QuantizedSeparableConv2DTranspose(Calibrable, SeparableConv2DTranspose):
         # depthwise process).
         inputs_channel_first = tf.transpose(inputs, (3, 0, 1, 2))
         inputs_channel_first = tf.expand_dims(inputs_channel_first, -1)
-        kernel_channel_first = tf.transpose(self.depthwise_kernel, (2, 0, 1, 3))
+        kernel_channel_first = tf.transpose(depthwise_kernel, (2, 0, 1, 3))
         kernel_channel_first = tf.expand_dims(kernel_channel_first, -2)
 
-        # Quantize the depthwise kernels
-        depthwise_kernel = self.dw_weight_quantizer(kernel_channel_first, training)
-
         # Perform the depthwise operation on values using conv2d_transpose on each channel
         dw_values = tf.vectorized_map(
             lambda x: backend.conv2d_transpose(x[0],
@@ -303,28 +303,23 @@ class QuantizedSeparableConv2DTranspose(Calibrable, SeparableConv2DTranspose):
                                                output_shape=output_shape,
                                                strides=self.strides,
                                                padding=self.padding),
-            (inputs_channel_first.values, depthwise_kernel.values))
+            (inputs_channel_first.values, kernel_channel_first.values))
         dw_values = tf.transpose(tf.squeeze(dw_values, axis=-1), (1, 2, 3, 0))
 
-        if isinstance(depthwise_kernel, QFloat):
-            # Multiply by the scale
-            dw_values *= depthwise_kernel.scales.values
-
         # Build a new FixedPoint
-        if isinstance(depthwise_kernel, FixedPoint):
-            filters_frac_bits = depthwise_kernel.frac_bits
-        else:
-            filters_frac_bits = depthwise_kernel.scales.frac_bits
-        dw_outputs = FixedPoint(dw_values, inputs.frac_bits + filters_frac_bits, inputs.value_bits)
+        dw_outputs = FixedPoint(dw_values, inputs.value_bits, inputs.frac_bits)
+        if isinstance(depthwise_kernel, QFloat):
+            # Build a new QFloat
+            dw_outputs = QFloat(dw_outputs, depthwise_kernel.scales)
 
         if self.intermediate_quantizer is not None:
-            dw_outputs_q = self.intermediate_quantizer(dw_outputs, training)
+            dw_outputs_q = self.intermediate_quantizer(dw_outputs)
             dw_outputs_q = dw_outputs_q.promote(self.buffer_bitwidth)
         else:
             dw_outputs_q = dw_outputs
 
         # Quantize the pointwise kernel
-        pointwise_kernel = self.pw_weight_quantizer(self.pointwise_kernel, training)
+        pointwise_kernel = self.pw_weight_quantizer(self.pointwise_kernel)
 
         # Pointwise operation
         outputs = tf.nn.convolution(
@@ -334,17 +329,13 @@ class QuantizedSeparableConv2DTranspose(Calibrable, SeparableConv2DTranspose):
             padding='VALID')
 
         if self.use_bias:
-            # Quantize biases
-            bias = self.bias_quantizer(self.bias, training)
-            # Align intermediate outputs and biases before adding them
-            outputs, shift = outputs.align(bias)
-            self.pw_shift(shift)
-            bias, shift = bias.promote(self.buffer_bitwidth).align(outputs)
-            self.bias_shift(shift)
+            # Quantize biases and align them on the outputs
+            bias = self.bias_quantizer(self.bias, outputs)
+            # Add biases
             outputs = tf.add(outputs, bias)
 
         if self.out_quantizer is not None:
-            outputs = self.out_quantizer(outputs, training)
+            outputs = self.out_quantizer(outputs)
         return outputs
 
     def get_config(self):
