@@ -17,9 +17,8 @@
 import keras
 import tensorflow as tf
 
-from .calibrable import Calibrable, CalibrableVariable
-from .quantizers import deserialize_quantizer, WeightQuantizer, OutputQuantizer
-from ..tensors import FixedPoint, MAX_BUFFER_BITWIDTH
+from .quantizers import AlignedWeightQuantizer, OutputQuantizer
+from ..tensors import QTensor, FixedPoint, MAX_BUFFER_BITWIDTH
 
 
 # Restrict list of exported symbols on default import
@@ -43,7 +42,7 @@ class ClassToken(keras.layers.Layer):
             trainable=True,
         )
 
-    def call(self, inputs, training=None):
+    def call(self, inputs):
         batch_size = tf.shape(inputs)[0]
         cls_broadcasted = tf.cast(
             tf.broadcast_to(self.cls, [batch_size, 1, self.hidden_size]),
@@ -68,57 +67,51 @@ class AddPositionEmbs(keras.layers.Layer):
             trainable=True,
         )
 
-    def call(self, inputs, training=None):
+    def call(self, inputs):
         return inputs + tf.cast(self.pe, dtype=inputs.dtype)
 
 
 @tf.keras.utils.register_keras_serializable()
-class QuantizedClassToken(Calibrable, ClassToken):
+class QuantizedClassToken(ClassToken):
     """Quantize the :class:`ClassToken` layer, allowing quantization of the output.
 
     Args:
-        quant_config (dict, optional): the serialized quantization configuration.
-            Defaults to empty configuration.
+        quant_config (dict): the serialized quantization configuration.
     """
 
-    def __init__(self, *args, quant_config={}, **kwargs):
+    def __init__(self, *args, quant_config, **kwargs):
         super().__init__(*args, **kwargs)
         self.quant_config = quant_config
-        self.out_quantizer = deserialize_quantizer(
-            self.quant_config, "output_quantizer", OutputQuantizer, False)
-        self.cls_quantizer = deserialize_quantizer(
-            self.quant_config, "cls_quantizer", WeightQuantizer, True)
+        out_quant_cfg = quant_config.get("output_quantizer", False)
+        if out_quant_cfg:
+            self.out_quantizer = OutputQuantizer(
+                name="output_quantizer", **out_quant_cfg)
+        else:
+            self.out_quantizer = None
+        self.cls_quantizer = AlignedWeightQuantizer(
+            name="cls_quantizer", **quant_config["cls_quantizer"])
         self.buffer_bitwidth = self.quant_config.get(
             "buffer_bitwidth", MAX_BUFFER_BITWIDTH) - 1
 
-        # Add objects that will store the shift values.
-        self.input_shift = CalibrableVariable()
-        self.cls_shift = CalibrableVariable()
-
-    def call(self, inputs, training=None):
-        # raise an error if the inputs are not FixedPoint
-        if not isinstance(inputs, FixedPoint):
-            raise TypeError(f"QuantizedClassToken only accepts FixedPoint\
+    def call(self, inputs):
+        # raise an error if the inputs are not QTensor
+        if not isinstance(inputs, QTensor):
+            raise TypeError(f"QuantizedClassToken only accepts QTensor\
                               inputs. Receives {type(inputs)} inputs.")
 
         batch_size = tf.shape(inputs.values)[0]
         inputs = inputs.promote(self.buffer_bitwidth)
 
-        cls = self.cls_quantizer(self.cls, training=training)
-        cls = cls.promote(self.buffer_bitwidth)
+        # Quantize the token and align it with the inputs
+        cls = self.cls_quantizer(self.cls, inputs)
 
+        # We need to broadcast the token along the batch size and set its token dimension to 1
         cls_broadcasted = tf.broadcast_to(cls, [batch_size, 1, self.hidden_size])
 
-        # Align inputs and cls_brodcasted frac_bits values before concatenation
-        inputs, shift = inputs.align(cls_broadcasted)
-        self.input_shift(shift)
-        cls_broadcasted, shift = cls_broadcasted.align(inputs)
-        self.cls_shift(shift)
-
         outputs = tf.concat([cls_broadcasted, inputs], 1)
 
         if self.out_quantizer is not None:
-            outputs = self.out_quantizer(outputs, training=training)
+            outputs = self.out_quantizer(outputs)
         return outputs
 
     def get_config(self):
@@ -128,42 +121,38 @@ class QuantizedClassToken(Calibrable, ClassToken):
 
 
 @tf.keras.utils.register_keras_serializable()
-class QuantizedAddPositionEmbs(Calibrable, AddPositionEmbs):
+class QuantizedAddPositionEmbs(AddPositionEmbs):
     """Quantize the :class:`AddPositionEmbs` layer, allowing operations in FixedPoint domain.
 
     Args:
-        quant_config (dict, optional): the serialized quantization configuration.
-            Defaults to empty configuration.
+        quant_config (dict): the serialized quantization configuration.
     """
 
-    def __init__(self, *args, quant_config={}, **kwargs):
+    def __init__(self, *args, quant_config, **kwargs):
         super().__init__(*args, **kwargs)
         self.quant_config = quant_config
-        self.out_quantizer = deserialize_quantizer(
-            self.quant_config, "output_quantizer", OutputQuantizer, False)
-        self.pe_quantizer = deserialize_quantizer(
-            self.quant_config, "pe_quantizer", WeightQuantizer, True)
+        out_quant_cfg = quant_config.get("output_quantizer", False)
+        if out_quant_cfg:
+            self.out_quantizer = OutputQuantizer(
+                name="output_quantizer", **out_quant_cfg)
+        else:
+            self.out_quantizer = None
+        self.pe_quantizer = AlignedWeightQuantizer(
+            name="pe_quantizer", **quant_config["pe_quantizer"])
         self.buffer_bitwidth = self.quant_config.get(
             "buffer_bitwidth", MAX_BUFFER_BITWIDTH) - 1
 
-        # Add objects that will store the shift values.
-        self.input_shift = CalibrableVariable()
-        self.pe_shift = CalibrableVariable()
-
-    def call(self, inputs, training=None):
-        # raise an error if the inputs are not FixedPoint
-        if not isinstance(inputs, FixedPoint):
-            raise TypeError(f"QuantizedAddPositionEmbs only accepts FixedPoint\
+    def call(self, inputs):
+        # raise an error if the inputs are not QTensor
+        if not isinstance(inputs, QTensor):
+            raise TypeError(f"QuantizedAddPositionEmbs only accepts QTensor\
                               inputs. Receives {type(inputs)} inputs.")
 
         inputs = inputs.promote(self.buffer_bitwidth)
-        pe = self.pe_quantizer(self.pe, training=training).promote(self.buffer_bitwidth)
+        # Quantize position embeddings and align them on the inputs
+        pe = self.pe_quantizer(self.pe, inputs)
 
-        # Align intermediate outputs and biases before adding them
-        inputs, shift = inputs.align(pe)
-        self.input_shift(shift)
-        pe, shift = pe.align(inputs)
-        self.pe_shift(shift)
+        # Add position embeddings
         outputs = inputs + pe
 
         if self.out_quantizer is not None:
@@ -189,7 +178,7 @@ class ExtractToken(keras.layers.Layer):
         super().__init__(*args, **kwargs)
         self.token = token
 
-    def call(self, inputs, training=False):
+    def call(self, inputs):
         return tf.gather(inputs, self.token, axis=1)
 
     def get_config(self):
@@ -203,7 +192,7 @@ class QuantizedExtractToken(ExtractToken):
     """ Quantized version of the ExtractToken layer. Accepts only FixedPoint inputs.
     """
 
-    def call(self, inputs, training=False):
+    def call(self, inputs):
         if not isinstance(inputs, FixedPoint):
             # If the inputs is not a FixedPoint, raise an error
             raise TypeError(f"QuantizedExtractToken only accepts FixedPoint\
