@@ -23,25 +23,30 @@ from keras import Model
 from .utils import deep_clone_model, insert_layer
 from .transforms import invert_batchnorm_pooling, fold_batchnorms
 from ..layers import (QuantizedDense, QuantizedLayerNormalization, QuantizedAddPositionEmbs,
-                      QuantizedClassToken, QuantizedGELU, QuantizedReLU, QuantizedAdd, QDropout,
+                      QuantizedClassToken, QuantizedReLU, QuantizedAdd, QDropout,
                       QuantizedReshape, QuantizedAttention, QuantizedConv2D, QuantizedMaxPool2D,
                       QuantizedGlobalAveragePooling2D, QuantizedExtractToken, Dequantizer,
-                      Quantizer, QuantizedSeparableConv2D, QuantizedFlatten, QuantizedReciprocal,
+                      QuantizedSeparableConv2D, QuantizedFlatten,
                       OutputQuantizer, WeightQuantizer, QuantizedConv2DTranspose,
-                      QuantizedShiftmax, QuantizedSeparableConv2DTranspose)
+                      QuantizedShiftmax, QuantizedSeparableConv2DTranspose,
+                      QuantizedDepthwiseConv2D, QuantizedUnfoldPatches, QuantizedFoldPatches,
+                      QuantizedConcatenate, QuantizedRescaling)
 
 # Mapper to match float layer with its quantized version
 CUSTOM_QLAYERS = {
     "Conv2D": QuantizedConv2D, "Dense": QuantizedDense, "Add": QuantizedAdd,
     "AddPositionEmbs": QuantizedAddPositionEmbs, "ClassToken": QuantizedClassToken,
-    "GELU": QuantizedGELU, "ReLU": QuantizedReLU, "Dropout": QDropout,
+    "ReLU": QuantizedReLU, "Dropout": QDropout,
     "LayerMadNormalization": QuantizedLayerNormalization, "Reshape": QuantizedReshape,
     "Attention": QuantizedAttention, 'MaxPooling2D': QuantizedMaxPool2D,
     'GlobalAveragePooling2D': QuantizedGlobalAveragePooling2D,
     "ExtractToken": QuantizedExtractToken, "SeparableConv2D": QuantizedSeparableConv2D,
-    "Flatten": QuantizedFlatten, "Reciprocal": QuantizedReciprocal, "PaddedConv2D": QuantizedConv2D,
+    "Flatten": QuantizedFlatten, "PaddedConv2D": QuantizedConv2D,
     "Conv2DTranspose": QuantizedConv2DTranspose, "Shiftmax": QuantizedShiftmax,
-    "SeparableConv2DTranspose": QuantizedSeparableConv2DTranspose
+    "SeparableConv2DTranspose": QuantizedSeparableConv2DTranspose,
+    "DepthwiseConv2D": QuantizedDepthwiseConv2D, "UnfoldPatches": QuantizedUnfoldPatches,
+    "FoldPatches": QuantizedFoldPatches, "Concatenate": QuantizedConcatenate,
+    "Rescaling": QuantizedRescaling
 }
 # List of Quantizer layer's that do not have a float layer representation
 NO_FLOAT_CUSTOM_QLAYERS = [Dequantizer, OutputQuantizer, WeightQuantizer]
@@ -252,7 +257,14 @@ def quantize(model, q_config, add_dequantizer=True):
         #    Cases covered in previous steps
 
         # If no match, return the qlayer with config unchanged
-        return get_quantize_layer(layer)
+        try:
+            qlayer = get_quantize_layer(layer)
+        except TypeError as e:
+            if "required keyword-only argument: 'quant_config'" in str(e):
+                raise TypeError(f"Make sure layer '{layer.name}' is not missing a configuration.") \
+                    from e
+            raise e
+        return qlayer
 
     # Invert BN <-> Pooling layers and fold BN into their preceding layers
     model = invert_batchnorm_pooling(model)
@@ -272,49 +284,16 @@ def quantize(model, q_config, add_dequantizer=True):
     return Model(new_model.input, out)
 
 
-def dump_config(model, skip_default=False):
+def dump_config(model):
     """Dump the quantization configuration of a quantized model, exporting the configuration for
     each quantized layer.
 
     Args:
         model (keras.Model): a quantized model.
-        skip_default (bool, optional): remove default values on each quantizer. Defaults to False.
 
     Returns:
         dict: the configuration of the model.
     """
-    # Get the simplest configuration by default of one quantizer
-    dump_obj = {"signed": True, "axis": "per-tensor", "bitwidth": None, "scale_bits": None}
-
-    # Function that update the configuration of each quantizer in one layer, reducing the
-    # parameters if short_form is True, otherwise adding the default parameters.
-    def _update_config(config):
-        for conf_value in config.values():
-            if isinstance(conf_value, dict):
-                # Check if current config is a singular quantizer
-                try:
-                    scale_bits = conf_value.pop("scale_bits", None)
-                    Quantizer(**conf_value)
-                    if scale_bits is not None:
-                        conf_value["scale_bits"] = scale_bits
-                except Exception:
-                    _update_config(conf_value)
-                    continue
-
-                # Remove extra keys
-                keys_to_pop = [key for key in conf_value.keys() if key not in dump_obj]
-                for ckey in keys_to_pop:
-                    conf_value.pop(ckey)
-
-                # Update the configuration of a singular quantizer follows:
-                # * skip_default=True: removes default values, given by Quantizer.get_config()
-                # * skip_default=False: keeps only [bitwidth, signed, axis, scale_bits] parameters
-                for dkey, dval in dump_obj.items():
-                    if not skip_default and dval is not None and dkey not in conf_value:
-                        conf_value[dkey] = dval
-                    elif skip_default and dval == conf_value.get(dkey, -1):
-                        conf_value.pop(dkey)
-
     # Get the configuration of the model, iterating over each layer and updating on config.
     config = {}
     for layer in model.layers:
@@ -323,7 +302,6 @@ def dump_config(model, skip_default=False):
 
         # Only append quantized configuration
         if layer.__class__ in CUSTOM_QLAYERS.values() and ly_config is not None:
-            _update_config(ly_config)
             config[layer.name] = ly_config
 
     return config
