@@ -16,28 +16,34 @@
 # ******************************************************************************
 import tensorflow as tf
 
-from .qtensor import QTensor, round_through, floor_through, pow2, ceil_log2
+from .qtensor import QTensor, round_through, floor_through, ceil_through, pow2, ceil_log2
 from ..debugging import assert_equal, assert_none_equal, assert_less, assert_less_equal
 
 
 class FixedPoint(QTensor):
     """A Tensor of integer values representing fixed-point numbers
 
+    The value_bits parameter sets the maximum integer values that can be stored:
+
+        int_max = 2^bits - 1.
+
+    When a QTensor is created, its values are clipped to [-int_max-1, int_max].
+
     Args:
         values (:obj:`tensorflow.Tensor`): a tensor of integer values
-        frac_bits (:obj:`tensorflow.Tensor`, optional): an integer tensor of
-            fractional bits. Defaults to 0.
-        value_bits (int, optional): the number of value bits. Defaults to 32.
+        value_bits (int): the number of value bits.
+        frac_bits (:obj:`tensorflow.Tensor`): an integer tensor of fractional bits.
     """
+    values: tf.Tensor = 1.0
+    value_bits: int = 7
     frac_bits: tf.Tensor = 0.
 
-    def __init__(self, values, frac_bits=0, value_bits=32):
+    def __init__(self, values, value_bits, frac_bits):
         # We store fractional bits in a float tensor to speed up calculations
         if isinstance(frac_bits, tf.Tensor):
             self.frac_bits = tf.cast(frac_bits, tf.float32)
         else:
             self.frac_bits = tf.convert_to_tensor(frac_bits, tf.float32)
-        assert_less_equal(0, self.frac_bits, "The frac_bits must be positive.")
         self.value_bits = value_bits
         # We store integer values in a float tensor to speed up calculations
         if isinstance(values, tf.Tensor):
@@ -48,12 +54,45 @@ class FixedPoint(QTensor):
         self.values = QTensor.clamp(values, value_bits)
         self.shape = self.values.shape
 
+    @property
+    def name(self):
+        return "no-name" if not hasattr(self.values, "name") else self.values.name
+
     @property
     def per_tensor(self):
-        return tf.rank(self.frac_bits) == 0
+        return self.frac_bits.shape.ndims == 0 or self.frac_bits.shape == (1,)
 
     @staticmethod
-    def quantize(x, value_bits, frac_bits=None):
+    def max_frac_bits(value_bits, ranges, clamp=True):
+        """Evaluate the maximum fractional bit index for the quantization ranges.
+
+        This method evaluates the minimum number of integer bits required to cover
+        the specified quantization ranges (this can be a negative number if the
+        ranges are strictly lower than zero).
+        From that it deduces the rightmost fractional bit indices. This can also be a negative
+        number if the ranges are higher than the biggest integer than can be represented with the
+        specified value bits)
+        If specified, it clamps the results to the available value_bits.
+
+        Args:
+            value_bits (int): the number of value bits.
+            ranges (:obj:`tensorflow:Tensor`): a tensor of float quantization ranges.
+            clamp (bool, optional): clamp the results to self.value_bits. Defaults to True.
+
+        Returns:
+            :obj:`tensorflow:Tensor`: a tensor of fractional bits.
+        """
+        # Get the powers of two containing the quantization ranges
+        int_bits = ceil_log2(ranges)
+        # Evaluate the rightmost fractional bits (they can be negative)
+        value_bits = tf.cast(value_bits, tf.float32)
+        frac_bits = value_bits - int_bits
+        if clamp:
+            frac_bits = tf.minimum(frac_bits, value_bits)
+        return frac_bits
+
+    @staticmethod
+    def quantize(x, value_bits, frac_bits=None, ceil=False):
         r"""Converts a float Tensor to a FixedPoint
 
         It converts the original float values into integer values so that:
@@ -85,6 +124,7 @@ class FixedPoint(QTensor):
             value_bits (int): the number of value bits
             frac_bits (:obj:`tensorflow.Tensor`, optional): an integer tensor of fractional bits.
                 Defaults to None.
+            ceil (bool, optional): ceil values instead of rounding them. Defaults to False.
 
         Returns:
             a :obj:`FixedPoint`
@@ -97,9 +137,11 @@ class FixedPoint(QTensor):
                 # Evaluate integer bits
                 int_bits = ceil_log2(tf.abs(x))
                 frac_bits = tf.maximum(0, value_bits - int_bits)
-        # Project float into integer fixed-point space
-        values = round_through(x * pow2(frac_bits))
-        return FixedPoint(values, frac_bits, value_bits)
+        # Project float into fixed-point representation space
+        x_scaled = x * pow2(frac_bits)
+        # Round or ceil to obtain integer values
+        values = ceil_through(x_scaled) if ceil else round_through(x_scaled)
+        return FixedPoint(values, value_bits, frac_bits)
 
     @property
     def sign(self):
@@ -108,7 +150,7 @@ class FixedPoint(QTensor):
         Returns:
             :obj:`FixedPoint`: the sign as a FixedPoint.
         """
-        return FixedPoint(tf.math.sign(self.values), 0, self.value_bits)
+        return FixedPoint(tf.math.sign(self.values), self.value_bits, 0)
 
     def to_float(self):
         return self.values / pow2(self.frac_bits)
@@ -129,71 +171,142 @@ class FixedPoint(QTensor):
         if bits < self.value_bits:
             raise ValueError(f"Cannot reduce value bits from {self.value_bits} to {bits}: "
                              "use a quantizer instead")
-        return FixedPoint(self.values, self.frac_bits, bits)
+        return FixedPoint(self.values, bits, self.frac_bits)
 
-    def align(self, other=None):
+    def align(self, other, value_bits=None):
         """Align fractional bits
 
         This returns an equivalent FixedPoint with a scalar fractional bit
-        corresponding to the maximum of:
-
-            - the initial fractional bits on all channels,
-            - when specified, the other FixedPoint fractional bits on all channels.
+        corresponding to the maximum of the current and other FixedPoint on all
+        channels.
 
         This is required before performing an operation that adds or subtracts
         elements along the last dimension, to make sure all these elements are
         in the same scale.
 
         Args:
-            other(:obj:`FixedPoint`): a FixedPoint to align to
+            other (:obj:`FixedPoint`): a FixedPoint to align to
+            value_bits (int, optional): the target value bits. Defaults to None.
 
         Returns:
             tuple(:obj:`FixedPoint`, :obj:`tf.Tensor`): a new FixedPoint with aligned
                 fractional bits and the shift that was applied.
         """
-        max_frac_bits = tf.reduce_max(self.frac_bits)
-        if other is not None:
-            if not isinstance(other, FixedPoint):
-                raise ValueError("Other must be a FixedPoint")
-            max_frac_bits = tf.math.maximum(max_frac_bits, tf.reduce_max(other.frac_bits))
-        shift = max_frac_bits - self.frac_bits
-        assert_less(shift, self.value_bits, "Cannot align the FixedPoint as it saturates its "
-                    f"{self.value_bits}-bit buffer.")
-        values = FixedPoint._lshift(self.values, shift)
-        return FixedPoint(values, max_frac_bits, self.value_bits), shift
+        if not isinstance(other, FixedPoint):
+            raise ValueError("Other must be a FixedPoint.")
+        max_frac_bits = tf.math.maximum(self.frac_bits, other.frac_bits)
+        return self.upscale(max_frac_bits, value_bits)
 
-    def downscale(self, frac_bits, value_bits):
-        """Encode a FixedPoint with a lower bitwidth
-
-        It applies a left/right shift to the source FixedPoint values to align
-        them on the target maximum fractional bits.
-
-        It then clips them to the boundaries specified by the target value bits.
+    def downscale(self, frac_bits):
+        """Reduce the precision of a FixedPoint
 
         Args:
             frac_bits (:obj:`tensorflow.Tensor`): the target fractional bits
-            value_bits (int): the target value bits
 
         Returns:
-            a :obj:`FixedPoint`
+            :obj:`FixedPoint`: the downscaled FixedPoint
         """
-        if value_bits > self.value_bits:
-            raise ValueError(
-                f"Cannot increase value bits from {self.value_bits} to {value_bits}:"
-                "use a promotion instead")
+        frac_bits = tf.cast(frac_bits, tf.float32)
+        assert_less_equal(frac_bits, self.frac_bits,
+                          f"Cannot reduce {self.name} precision because the target precision "
+                          f"({frac_bits}) is higher than the current one ({self.frac_bits})")
+        return self.rescale(frac_bits)
+
+    def rescale(self, frac_bits, value_bits=None):
+        """Rescale a FixedPoint to a specified precision and bitwidth
+
+        This primarily rescales the FixedPoint values to match the precision
+        specified by the target fractional bits.
+
+        Optionally, this adjusts the value bits to the specified bitwidth.
+
+        The rescaling operation is:
+
+        - a left shift of the values when their precision increases,
+        - a rounded right shift of the values when their precision decreases.
+
+        This method can be used to:
+
+        - compress a FixedPoint to a lower bitwidth after having reduced its precision,
+        - expand a FixedPoint to a larger bitwidth after having increased its precision.
+
+        Args:
+            frac_bits (:obj:`tensorflow.Tensor`): the target fractional bits
+            value_bits (int, optional): the target value bits
+
+        Returns:
+            :obj:`FixedPoint`: the rescaled FixedPoint
+        """
+        if value_bits is None:
+            value_bits = self.value_bits
         frac_bits = tf.cast(frac_bits, tf.float32)
         # Evaluate the shift to apply to reach the target precision
         shift = frac_bits - self.frac_bits
-        assert_less(tf.abs(shift), self.value_bits, "Cannot downscale the FixedPoint as it "
-                    f"saturates its {self.value_bits}-bit buffer.")
-        # The shift can be positive (left-shift) or negative (right-shift)
-        # In float it can be simulated by flooring after multiplying by the
-        # positive/negative shift PoT:
-        # - for positive shift, the results are integer and can safely be floored,
-        # - for negative shift, flooring simulates a right_shift.
-        values = round_through(self.values * pow2(shift))
+        # The shift can be positive (left-shift) or negative (rounded right-shift)
+        # A positive shift exceeding the target bitwidth always leads to a saturation
+        assert_less(shift, value_bits,
+                    f"Cannot rescale {self.name} to {frac_bits} fractional bits as it will saturate"
+                    f" its {value_bits} buffer")
+        # The integer operations are simulated in float arithmetics:
+        # - the shifts by multiplying by the positive/negative shift power of two,
+        # - the rounding by adding 0.5 and flooring (to workaround banker's rounding).
+        # For positive shift, the results being integer, the rounding has no effect.
+        # We therefore apply the same operations for both shifts.
+        values = floor_through(self.values * pow2(shift) + 0.5)
         # return a new FixedPoint with the target precision and bitwidth
-        return FixedPoint(values, frac_bits, value_bits), shift
+        return FixedPoint(values, value_bits, frac_bits), shift
+
+    def upscale(self, frac_bits, value_bits=None):
+        """Align a FixedPoint to a specified precision
+
+        The target precision must be higher than the current one.
+
+        Args:
+            frac_bits (:obj:`tensorflow.Tensor`): the target fractional bits
+            value_bits (int, optional): the target value bits
+
+        Returns:
+            :obj:`FixedPoint`: the upscaled FixedPoint
+        """
+        if value_bits is None:
+            value_bits = self.value_bits
+        frac_bits = tf.cast(frac_bits, tf.float32)
+        assert_less_equal(self.frac_bits, frac_bits,
+                          f"Cannot increase {self.name} precision because the target precision "
+                          f"({frac_bits}) is lower than the current one ({self.frac_bits})")
+        # Evaluate the shift to apply to reach the target precision
+        shift = frac_bits - self.frac_bits
+        assert_less(shift, value_bits,
+                    f"Cannot upscale {self.name} to {frac_bits} fractional bits as it saturates its"
+                    f" {value_bits}-bit buffer.")
+        # Apply shift
+        values = self.values * pow2(shift)
+        # return a new FixedPoint with the target precision and bitwidth
+        return FixedPoint(values, value_bits, frac_bits), shift
+
+    def expand(self, value_bits):
+        """Expand the FixedPoint to the specified bitwidth
+
+        This returns an equivalent FixedPoint with a higher or equal number of
+        value bits and a scalar fractional bit corresponding to the maximum of
+        the initial fractional bits on all channels.
+
+        This is mostly used to recover a per-tensor FixedPoint that has been
+        compressed to a lower number of value bits.
+
+        Args:
+            value_bits (int): the target value_bits
+
+        Returns:
+            tuple(:obj:`FixedPoint`, :obj:`tf.Tensor`): a new FixedPoint with expanded
+                fractional bits and the shift that was applied.
+        """
+        if value_bits < self.value_bits:
+            raise ValueError(
+                f"Cannot reduce {self.name} bitwidth from {self.value_bits} to {value_bits}:"
+                " use a quantizer instead.")
+        max_frac_bits = tf.reduce_max(self.frac_bits)
+        return self.upscale(max_frac_bits, value_bits)
 
     @staticmethod
     def _rshift(values, shift):
@@ -225,7 +338,7 @@ class FixedPoint(QTensor):
             :obj:`FixedPoint`: the result as a FixedPoint
         """
         values = floor_through(self.values * pow2(s))
-        return FixedPoint(values, self.frac_bits, self.value_bits)
+        return FixedPoint(values, self.value_bits, self.frac_bits)
 
     def __rshift__(self, shift):
         """Right shift the FixedPoint values
@@ -253,7 +366,7 @@ class FixedPoint(QTensor):
         s_values = self.values
         # Return a new FixedPoint with updated fractional bits,
         # which is equivalent in hardward without performing any operation
-        return FixedPoint(s_values, s_frac_bits, self.value_bits)
+        return FixedPoint(s_values, self.value_bits, s_frac_bits)
 
     def __lshift__(self, shift):
         """Left shift the FixedPoint values
@@ -272,7 +385,7 @@ class FixedPoint(QTensor):
         # Simply apply the shift on the values
         s_values = FixedPoint._lshift(self.values, shift)
         # Return a new FixedPoint with updated values
-        return FixedPoint(s_values, self.frac_bits, self.value_bits)
+        return FixedPoint(s_values, self.value_bits, self.frac_bits)
 
     def _align_values(self, other):
         # The sub fractional bits are the max of both terms
@@ -289,13 +402,11 @@ class FixedPoint(QTensor):
             # aligned with the current FixedPoint
             return self + FixedPoint.quantize(other, 32, self.frac_bits)
         elif isinstance(other, FixedPoint):
-            # Check that self and other are per-tensor
-            self.assert_per_tensor()
-            other.assert_per_tensor()
             # Check that self and other are aligned
-            assert_equal(self.frac_bits, other.frac_bits)
+            assert_equal(self.frac_bits, other.frac_bits,
+                         message=f"{self.name} and {other.name} are not aligned")
             # Return a new FixedPoint
-            return FixedPoint(self.values + other.values, self.frac_bits, self.value_bits)
+            return FixedPoint(self.values + other.values, self.value_bits, self.frac_bits)
         raise TypeError(
             f"Unsupported operand type(s) for +: 'FixedPoint' and '{type(other)}'")
 
@@ -305,13 +416,11 @@ class FixedPoint(QTensor):
             # aligned with the current FixedPoint
             return self - FixedPoint.quantize(other, 32, self.frac_bits)
         elif isinstance(other, FixedPoint):
-            # Check that self and other are per-tensor
-            self.assert_per_tensor()
-            other.assert_per_tensor()
             # Check that self and other are aligned
-            assert_equal(self.frac_bits, other.frac_bits)
+            assert_equal(self.frac_bits, other.frac_bits,
+                         message=f"{self.name} and {other.name} are not aligned")
             # Return a new FixedPoint
-            return FixedPoint(self.values - other.values, self.frac_bits, self.value_bits)
+            return FixedPoint(self.values - other.values, self.value_bits, self.frac_bits)
         raise TypeError(
             f"Unsupported operand type(s) for -: 'FixedPoint' and '{type(other)}'")
 
@@ -322,28 +431,39 @@ class FixedPoint(QTensor):
             # The product between fixed-point is straight-forward
             p_values = self.values * other.values
             # Return a new FixedPoint whose frac bits is the sum of both terms
-            return FixedPoint(p_values, self.frac_bits + other.frac_bits, self.value_bits)
+            return FixedPoint(p_values, self.value_bits, self.frac_bits + other.frac_bits)
         raise TypeError(
             f"Unsupported operand type(s) for *: 'FixedPoint' and '{type(other)}'")
 
     def __truediv__(self, other):
+        @tf.function
+        @tf.custom_gradient
+        def truncate(x):
+            # Remove decimal part from x, This is to obtain a value that
+            # matches the result estimated in C when working with ints.
+            rounded = tf.cast(tf.cast(x, tf.int32), tf.float32)
+
+            def grad(upstream):
+                return upstream
+            return rounded, grad
+
         if isinstance(other, int):
-            return self / FixedPoint(other)
+            return self / FixedPoint(other, 32, 0)
         elif isinstance(other, FixedPoint):
             assert_none_equal(other.values, 0.0, "Cannot divide by 0.")
             # The division between fixed-point is straight-forward
-            d_values = floor_through(self.values / other.values)
+            d_values = truncate(self.values / other.values)
             # Return a new FixedPoint whose frac bits from other is subtracted.
-            return FixedPoint(d_values, self.frac_bits - other.frac_bits, self.value_bits)
+            return FixedPoint(d_values, self.value_bits, self.frac_bits - other.frac_bits)
         raise TypeError(
             f"Unsupported operand type(s) for /: 'FixedPoint' and '{type(other)}'")
 
     def __pow__(self, power):
         if isinstance(power, int):
             if power == 0:
-                return FixedPoint(tf.ones_like(self.values), 0, self.value_bits)
+                return FixedPoint(tf.ones_like(self.values), self.value_bits, 0)
             elif power == 1:
-                return FixedPoint(self.values, self.frac_bits, self.value_bits)
+                return FixedPoint(self.values, self.value_bits, self.frac_bits)
             elif power > 1:
                 return self * self ** (power - 1)
             else:
@@ -400,7 +520,7 @@ class FixedPoint(QTensor):
         Returns:
             :obj:`FixedPoint`: the absolute value.
         """
-        return FixedPoint(tf.math.abs(self.values), self.frac_bits, self.value_bits)
+        return FixedPoint(tf.math.abs(self.values), self.value_bits, self.frac_bits)
 
     def floor(self):
         """Floors the FixedPoint
@@ -411,4 +531,4 @@ class FixedPoint(QTensor):
         """
         # Divide values to remove fractional bits
         values = FixedPoint._rshift(self.values, self.frac_bits)
-        return FixedPoint(values, 0, self.value_bits), self.frac_bits
+        return FixedPoint(values, self.value_bits, 0), self.frac_bits
